import unittest

from mygrations.formats.mysql.file_reader.parsers.constraint_foreign import constraint_foreign
class test_constraint_foreign(unittest.TestCase):
    def test_simple(self):

        # parse a typical foreign key constraint
        parser = constraint_foreign()
        returned = parser.parse(
            'CONSTRAINT `accounts_status_id_ref_account_statuses_id` FOREIGN KEY (`status_id`) REFERENCES `account_statuses` (`id`) ON DELETE CASCADE ON UPDATE SET NULL'
        )

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have matched everything
        self.assertEquals('', returned)

        # we should have lots of data now
        self.assertEquals('accounts_status_id_ref_account_statuses_id', parser.name)
        self.assertEquals('status_id', parser.column)
        self.assertEquals('account_statuses', parser.foreign_table)
        self.assertEquals('id', parser.foreign_column)
        self.assertEquals('CASCADE', parser.on_delete)
        self.assertEquals('SET NULL', parser.on_update)
        self.assertFalse(parser.has_comma)
        self.assertEquals(
            'CONSTRAINT `accounts_status_id_ref_account_statuses_id` FOREIGN KEY (`status_id`) REFERENCES `account_statuses` (`id`) ON DELETE CASCADE ON UPDATE SET NULL',
            str(parser)
        )

    def test_all_deletes(self):

        # parse a typical foreign key constraint
        parser = constraint_foreign()
        parser.parse('CONSTRAINT blah FOREIGN KEY (check) REFERENCES tbl (id) ON DELETE CASCADE ON UPDATE SET NULL')
        self.assertEquals('CASCADE', parser.on_delete)
        self.assertEquals(
            'CONSTRAINT `blah` FOREIGN KEY (`check`) REFERENCES `tbl` (`id`) ON DELETE CASCADE ON UPDATE SET NULL',
            str(parser)
        )

        # parse a typical foreign key constraint
        parser = constraint_foreign()
        parser.parse('CONSTRAINT blah FOREIGN KEY (check) REFERENCES tbl (id) ON DELETE NO ACTION ON UPDATE SET NULL')
        self.assertEquals('NO ACTION', parser.on_delete)
        self.assertEquals(
            'CONSTRAINT `blah` FOREIGN KEY (`check`) REFERENCES `tbl` (`id`) ON DELETE NO ACTION ON UPDATE SET NULL',
            str(parser)
        )

        # parse a typical foreign key constraint
        parser = constraint_foreign()
        parser.parse('CONSTRAINT blah FOREIGN KEY (check) REFERENCES tbl (id) ON DELETE RESTRICT ON UPDATE SET NULL')
        self.assertEquals('RESTRICT', parser.on_delete)
        self.assertEquals(
            'CONSTRAINT `blah` FOREIGN KEY (`check`) REFERENCES `tbl` (`id`) ON DELETE RESTRICT ON UPDATE SET NULL',
            str(parser)
        )

        # parse a typical foreign key constraint
        parser = constraint_foreign()
        parser.parse('CONSTRAINT blah FOREIGN KEY (check) REFERENCES tbl (id) ON DELETE set default ON UPDATE SET NULL')
        self.assertEquals('SET DEFAULT', parser.on_delete)
        self.assertEquals(
            'CONSTRAINT `blah` FOREIGN KEY (`check`) REFERENCES `tbl` (`id`) ON DELETE SET DEFAULT ON UPDATE SET NULL',
            str(parser)
        )

        # parse a typical foreign key constraint
        parser = constraint_foreign()
        parser.parse('CONSTRAINT blah FOREIGN KEY (check) REFERENCES tbl (id) ON DELETE set null ON UPDATE SET NULL')
        self.assertEquals('SET NULL', parser.on_delete)
        self.assertEquals(
            'CONSTRAINT `blah` FOREIGN KEY (`check`) REFERENCES `tbl` (`id`) ON DELETE SET NULL ON UPDATE SET NULL',
            str(parser)
        )

        # parse a typical foreign key constraint
        parser = constraint_foreign()
        parser.parse('CONSTRAINT blah FOREIGN KEY (check) REFERENCES tbl (id) ON UPDATE SET NULL')
        self.assertEquals('RESTRICT', parser.on_delete)
        self.assertEquals(
            'CONSTRAINT `blah` FOREIGN KEY (`check`) REFERENCES `tbl` (`id`) ON DELETE RESTRICT ON UPDATE SET NULL',
            str(parser)
        )

    def test_all_updates(self):

        # parse a typical foreign key constraint
        parser = constraint_foreign()
        parser.parse('CONSTRAINT blah FOREIGN KEY (check) REFERENCES tbl (id) ON DELETE CASCADE ON UPDATE SET NULL')
        self.assertEquals('SET NULL', parser.on_update)
        self.assertEquals(
            'CONSTRAINT `blah` FOREIGN KEY (`check`) REFERENCES `tbl` (`id`) ON DELETE CASCADE ON UPDATE SET NULL',
            str(parser)
        )

        # parse a typical foreign key constraint
        parser = constraint_foreign()
        parser.parse(
            'CONSTRAINT blah FOREIGN KEY (check) REFERENCES tbl (id) ON DELETE NO ACTION ON UPDATE SET DEFAULT'
        )
        self.assertEquals('SET DEFAULT', parser.on_update)
        self.assertEquals(
            'CONSTRAINT `blah` FOREIGN KEY (`check`) REFERENCES `tbl` (`id`) ON DELETE NO ACTION ON UPDATE SET DEFAULT',
            str(parser)
        )

        # parse a typical foreign key constraint
        parser = constraint_foreign()
        parser.parse('CONSTRAINT blah FOREIGN KEY (check) REFERENCES tbl (id) ON DELETE RESTRICT ON UPDATE CASCADE')
        self.assertEquals('CASCADE', parser.on_update)
        self.assertEquals(
            'CONSTRAINT `blah` FOREIGN KEY (`check`) REFERENCES `tbl` (`id`) ON DELETE RESTRICT ON UPDATE CASCADE',
            str(parser)
        )

        # parse a typical foreign key constraint
        parser = constraint_foreign()
        parser.parse(
            'CONSTRAINT blah FOREIGN KEY (check) REFERENCES tbl (id) ON DELETE set default ON UPDATE no action'
        )
        self.assertEquals('NO ACTION', parser.on_update)
        self.assertEquals(
            'CONSTRAINT `blah` FOREIGN KEY (`check`) REFERENCES `tbl` (`id`) ON DELETE SET DEFAULT ON UPDATE NO ACTION',
            str(parser)
        )

        # parse a typical foreign key constraint
        parser = constraint_foreign()
        parser.parse('CONSTRAINT blah FOREIGN KEY (check) REFERENCES tbl (id) ON DELETE set null ON UPDATE RESTRICT')
        self.assertEquals('RESTRICT', parser.on_update)
        self.assertEquals(
            'CONSTRAINT `blah` FOREIGN KEY (`check`) REFERENCES `tbl` (`id`) ON DELETE SET NULL ON UPDATE RESTRICT',
            str(parser)
        )

        # parse a typical foreign key constraint
        parser = constraint_foreign()
        parser.parse('CONSTRAINT blah FOREIGN KEY (check) REFERENCES tbl (id) ON DELETE SET NULL')
        self.assertEquals('RESTRICT', parser.on_update)
        self.assertEquals(
            'CONSTRAINT `blah` FOREIGN KEY (`check`) REFERENCES `tbl` (`id`) ON DELETE SET NULL ON UPDATE RESTRICT',
            str(parser)
        )

    def test_action_optional(self):

        # all actions are optional, and default to RESTRICT
        parser = constraint_foreign()
        remaining = parser.parse('CONSTRAINT blah FOREIGN KEY (check) REFERENCES tbl (id)')

        self.assertTrue(parser.matched)
        self.assertEquals('', remaining)
        self.assertEquals('RESTRICT', parser.on_update)
        self.assertEquals('RESTRICT', parser.on_delete)
        self.assertEquals(
            'CONSTRAINT `blah` FOREIGN KEY (`check`) REFERENCES `tbl` (`id`) ON DELETE RESTRICT ON UPDATE RESTRICT',
            str(parser)
        )

    def test_leftovers(self):

        # anything that doesn't match should be returned as leftovers
        parser = constraint_foreign()
        remaining = parser.parse('CONSTRAINT blah FOREIGN KEY (check) REFERENCES tbl (id), sup')

        self.assertTrue(parser.matched)
        self.assertEquals('sup', remaining)
        self.assertTrue(parser.has_comma)
        self.assertEquals(
            'CONSTRAINT `blah` FOREIGN KEY (`check`) REFERENCES `tbl` (`id`) ON DELETE RESTRICT ON UPDATE RESTRICT',
            str(parser)
        )

        # even if the user forgets a comma
        parser = constraint_foreign()
        remaining = parser.parse('CONSTRAINT blah FOREIGN KEY (check) REFERENCES tbl (id) sup')
        self.assertEquals(
            'CONSTRAINT `blah` FOREIGN KEY (`check`) REFERENCES `tbl` (`id`) ON DELETE RESTRICT ON UPDATE RESTRICT',
            str(parser)
        )

        self.assertTrue(parser.matched)
        self.assertEquals('sup', remaining)
        self.assertFalse(parser.has_comma)



class row_delete:
    """ Generates an SQL command to delete a record """

    def __init__(self, table_name, row_id):
        if type(table_name) != str:
            self._table_name = table_name.name
        else:
            self._table_name = table_name
        self.row_id = row_id

    @property
    def table_name(self):
        """ Public getter.  Returns the name of the table.

        :returns: The table name
        :rtype: string
        """

        return self._table_name

    def __str__(self):
        return 'DELETE FROM `%s` WHERE id=%s;' % (self._table_name, self.row_id)



class constraint(object):

    _errors = None
    _warnings = None
    _name = ''
    _column = ''
    _foreign_table = ''
    _foreign_column = ''
    _on_delete = ''
    _on_update = ''

    def __init__(self, name='', column='', foreign_table='', foreign_column='', on_delete='', on_update=''):
        self._name = name
        self._column = column
        self._foreign_table = foreign_table
        self._foreign_column = foreign_column
        self._on_delete = on_delete
        self._on_update = on_update

    @property
    def name(self):
        """ Public getter.  Returns the name of the column.

        :returns: The index name
        :rtype: string
        """

        return self._name

    @property
    def column(self):
        """ Public getter.  Returns the name of the column this constraint is on

        :returns: The column name
        :rtype: string
        """

        return self._column

    @property
    def foreign_table(self):
        """ Public getter.  Returns the name of the table this constraint is to

        :returns: The table name
        :rtype: string
        """

        return self._foreign_table

    @property
    def foreign_column(self):
        """ Public getter.  Returns the name of the column in the foreign table this constraint is to

        :returns: The column name
        :rtype: string
        """

        return self._foreign_column

    @property
    def on_delete(self):
        """ Public getter.  Returns the ON DELETE action for this constraint

        :returns: The ON DELETE action
        :rtype: string
        """

        return self._on_delete

    @property
    def on_update(self):
        """ Public getter.  Returns the ON UPDATE action for this constraint

        :returns: The ON UPDATE action
        :rtype: string
        """

        return self._on_update

    @property
    def errors(self):
        """ Public getter.  Returns a list of parsing errors

        :returns: A list of parsing errors
        :rtype: list
        """
        return [] if self._errors is None else self._errors

    @property
    def warnings(self):
        """ Public getter.  Returns a list of parsing/table warnings

        :returns: A list of parsing/table warnings
        :rtype: list
        """
        return [] if self._warnings is None else self._warnings

    def __str__(self):
        """ Returns the MySQL command that would create the constraint

        i.e. CONSTRAINT `vendors_w9_fk` FOREIGN KEY (`w9_id`) REFERENCES `vendor_w9s` (`id`) ON UPDATE CASCADE

        :returns: A partial MySQL command that could be used to generate the foreign key
        :rtype: string
        """
        parts = ['CONSTRAINT']

        parts.append('`%s`' % self.name)
        parts.append('FOREIGN KEY')
        parts.append('(`%s`)' % (self.column))
        parts.append('REFERENCES')
        parts.append('`%s`' % self.foreign_table)
        parts.append('(`%s`)' % (self.foreign_column))
        parts.append('ON DELETE %s' % self.on_delete)
        parts.append('ON UPDATE %s' % self.on_update)

        return ' '.join(parts)



import unittest

from mygrations.formats.mysql.db_reader.database import database as database_reader
from mygrations.drivers.mysqldb.mysqldb import mysqldb
from tests.mocks.db.mysql.db_structure import db_structure
from mygrations.formats.mysql.mygrations.row_mygration import row_mygration
class test_mygrate_rows_simple(unittest.TestCase):
    def _get_tables(self):

        return {
            'logs':
            """
                CREATE TABLE `logs` (
                    `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                    `message` TEXT NOT NULL,
                    `traceback` text,
                    PRIMARY KEY (`id`)
                ) ENGINE=InnoDB DEFAULT CHARSET=utf8;""",
            'more_logs':
            """
                CREATE TABLE `more_logs` (
                    `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                    `more_messages` TEXT NOT NULL,
                    `traceback` text,
                    PRIMARY KEY (`id`)
                ) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        }

    def test_add_rows(self):

        # I could also just create the database and add rows but I feel like doing this
        tables = self._get_tables()
        rows = {
            'logs': ({
                'id': 1,
                'message': 'hey',
                'traceback': 'never'
            }, {
                'id': 2,
                'message': 'sup',
                'traceback': 'forever'
            })
        }

        database = database_reader(mysqldb(db_structure(tables, rows)))
        database.read_rows('logs')

        operations = [str(op) for op in row_mygration(database).operations]
        self.assertEquals(2, len(operations))
        self.assertEquals(
            "INSERT INTO `logs` (`id`, `message`, `traceback`) VALUES ('1', 'hey', 'never');", operations[0]
        )
        self.assertEquals(
            "INSERT INTO `logs` (`id`, `message`, `traceback`) VALUES ('2', 'sup', 'forever');", operations[1]
        )

    def test_modify_rows(self):

        # I could also just create the database and add rows but I feel like doing this
        tables = self._get_tables()
        rows = {
            'logs': ({
                'id': 1,
                'message': 'hey',
                'traceback': 'never'
            }, {
                'id': 2,
                'message': 'sup',
                'traceback': 'forever'
            })
        }

        database = database_reader(mysqldb(db_structure(tables, rows)))
        database.read_rows('logs')

        rows_from = {
            'logs': ({
                'id': 1,
                'message': 'hey',
                'traceback': 'bah'
            }, {
                'id': 2,
                'message': 'sup',
                'traceback': 'forever'
            })
        }
        database_from = database_reader(mysqldb(db_structure(tables, rows_from)))
        database_from.read_rows('logs')

        operations = [str(op) for op in row_mygration(database, database_from).operations]
        self.assertEquals(1, len(operations))
        self.assertEquals("UPDATE `logs` SET `message`='hey', `traceback`='never' WHERE id=1;", operations[0])

    def test_delete_rows(self):

        # I could also just create the database and add rows but I feel like doing this
        tables = self._get_tables()
        rows = {'logs': ({'id': 2, 'message': 'sup', 'traceback': 'forever'}, )}

        database = database_reader(mysqldb(db_structure(tables, rows)))
        database.read_rows('logs')

        rows_from = {
            'logs': ({
                'id': 1,
                'message': 'hey',
                'traceback': 'never'
            }, {
                'id': 2,
                'message': 'sup',
                'traceback': 'forever'
            })
        }
        database_from = database_reader(mysqldb(db_structure(tables, rows_from)))
        database_from.read_rows('logs')

        operations = [str(op) for op in row_mygration(database, database_from).operations]
        self.assertEquals(1, len(operations))
        self.assertEquals("DELETE FROM `logs` WHERE id=1;", operations[0])

    def test_all(self):

        # I could also just create the database and add rows but I feel like doing this
        tables = self._get_tables()
        rows = {
            'logs': ({
                'id': 2,
                'message': 'sup',
                'traceback': 'whatever'
            }, {
                'id': 3,
                'message': 'okay',
                'traceback': 'always'
            })
        }

        database = database_reader(mysqldb(db_structure(tables, rows)))
        database.read_rows('logs')

        rows_from = {
            'logs': ({
                'id': 1,
                'message': 'hey',
                'traceback': 'never'
            }, {
                'id': 2,
                'message': 'sup',
                'traceback': 'forever'
            })
        }
        database_from = database_reader(mysqldb(db_structure(tables, rows_from)))
        database_from.read_rows('logs')

        operations = [str(op) for op in row_mygration(database, database_from).operations]
        # don't be picky about the order
        self.assertEquals(3, len(operations))
        self.assertTrue(
            "INSERT INTO `logs` (`id`, `message`, `traceback`) VALUES ('3', 'okay', 'always');" in operations
        )
        self.assertTrue("DELETE FROM `logs` WHERE id=1;" in operations)
        self.assertTrue("UPDATE `logs` SET `message`='sup', `traceback`='whatever' WHERE id=2;" in operations)



class change_key:
    """ Generates a partial SQL command to change a key in a table """

    def __init__(self, key):
        self.key = key

    def apply_to_table(self, table):
        """ Changes the key in the table

        :param table: The table to change the key on
        :param type: mygrations.formats.mysql.definitions.table
        """
        table.change_key(self.key)

    def __str__(self):
        return 'DROP KEY `%s`, ADD %s' % (self.key.name, str(self.key))
